# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017
"""
Reading and writing of files.
"""

import os
import enum
import streamsx.spl.op
from streamsx.topology.schema import StreamSchema
from streamsx.standard import CloseMode, Format, Compression, WriteFailureAction
import streamsx.topology.composite

import streamsx.standard._version
__version__ = streamsx.standard._version.__version__


class FileSink(streamsx.topology.composite.ForEach):
    """
    Write a stream to a file

    Example for writing a stream to a file::

        import streamsx.standard.files as files
        from streamsx.topology.topology import Topology

        topo = Topology()
        s = topo.source(['Hello', 'World!']).as_string()
        s.for_each(files.FileSink(file=os.path.join(tempfile.mkdtemp(), 'data.txt')))

    .. versionadded:: 0.5

    Attributes
    ----------
    file : str
        Name of the output file.
    """


    def __init__(self, file):
        self.file = file
        self.append = None
        self.bytes_per_file = None
        self.close_mode = None
        self.compression = None
        self.encoding = None
        self.eol_marker = None
        self.flush = None
        self.flush_on_punctuation = None
        self.format = None
        self.has_delay_field = None
        self.move_file_to_directory = None
        self.quote_strings = None
        self.separator = None
        self.suppress = None
        self.time_per_file = None
        self.truncate_on_reset = None
        self.tuples_per_file = None
        self.write_failure_action = None
        self.write_punctuations = None

    @property
    def append(self):
        """
            bool: Specifies that the generated tuples are appended to the output file. If this parameter is false or not specified, the output file is truncated before the tuples are generated.
        """
        return self._append

    @append.setter
    def append(self, value):
        self._append = value


    @property
    def bytes_per_file(self):
        """
            int: Specifies the approximate size of the output file, in bytes. When the file size exceeds the specified number of bytes, the current output file is closed and a new file is opened. This parameter must be specified when the :py:meth:`~streamsx.standard.files.FileSink.close_mode` parameter is set to size.
        """
        return self._bytes_per_file

    @bytes_per_file.setter
    def bytes_per_file(self, value):
        self._bytes_per_file = value

    @property
    def close_mode(self):
        """
            enum: Specifies when the file closes and a new one opens. This parameter has type enum {punct, count, size, time, dynamic, never} :py:meth:`~streamsx.standard.CloseMode`. The default value is never. For any other value except dynamic, when the specified condition is satisfied, the current output file is closed and a new file is opened for writing.
The parameter value:
punct specifies to close the file when a window or final punctuation is received.
count is used with the :py:meth:`~streamsx.standard.files.FileSink.tuples_per_file` parameter to close the file when the specified number of tuples have been received.
size is used with the :py:meth:`~streamsx.standard.files.FileSink.bytes_per_file` parameter to close the file when the specified number of bytes have been received.
time is used with the :py:meth:`~streamsx.standard.files.FileSink.time_per_file` parameter to close the file when the specified time has elapsed.
If this parameter value is dynamic, the file parameter can reference input attributes and is evaluated at each input tuple to compute the file name. If the file name is different from the previous value, the output file closes, and a new file opens.
In all cases, the file parameter can contain modifiers that are used to generate the file name to be used. The supported modifiers are:
{id}: Each {id} in the file name is replaced with the file number created by the FileSink operator. It has value 0 for the first file, 1 for the second file, and so on.
{localtime:strftimeString}: The contents are replaced by the current local time, formatted as if by the strftime library call.
{gmtime:strftimeString}: The contents are replaced by the current time in the GMT timezone. They are formatted as if by the strftime library call.
The modifiers can be repeated in the string, and are all replaced with their values. If :py:meth:`~streamsx.standard.files.FileSink.close_mode` is dynamic, the file names are compared after the modifiers are substituted.
        """
        return self._close_mode

    @close_mode.setter
    def close_mode(self, value):
        self._close_mode = value

    @property
    def compression(self):
        """
             enum: Specifies the compression mode :py:meth:`~streamsx.standard.Compression`
        """
        return self._compression

    @compression.setter
    def compression(self, value):
        self._compression = value

    @property
    def encoding(self):
        """
             str: Specifies the character set encoding that is used in the output file. Data that is written to the output file is converted from the UTF-8 character set to the specified character set before any compression is performed. The encoding parameter is not valid with formats bin or block.
        """
        return self._encoding

    @encoding.setter
    def encoding(self, value):
        self._encoding = value

    @property
    def eol_marker(self):
        """
             str: Specifies the end of line marker.
        """
        return self._eol_marker

    @eol_marker.setter
    def eol_marker(self, value):
        self._eol_marker = value

    @property
    def flush(self):
        """
            int: Specifies the number of tuples after which to flush the output file. By default no flushing on tuple numbers is performed. 
        """
        return self._flush

    @flush.setter
    def flush(self, value):
        self._flush = value

    @property
    def flush_on_punctuation(self):
        """
            bool: Specifies to flush the output file when a window or final punctuation is received. This parameter defaults to true. 
        """
        return self._flush_on_punctuation

    @flush_on_punctuation.setter
    def flush_on_punctuation(self, value):
        self._flush_on_punctuation = value

    @property
    def format(self):
        """
            enum: Specifies the format of the data :py:meth:`~streamsx.standard.Format`
        """
        return self._format

    @format.setter
    def format(self, value):
        self._format = value

    @property
    def has_delay_field(self):
        """
            bool: Specifies whether to output an extra attribute per tuple, which specifies the inter-arrival delays between the input tuples
        """
        return self._has_delay_field

    @has_delay_field.setter
    def has_delay_field(self, value):
        self._has_delay_field = value

    @property
    def move_file_to_directory(self):
        """
            str: Specifies that the file is moved to the named directory after the file is closed. Any existing file with same name is removed before the file is moved to the move_file_to_directory directory.
        """
        return self._move_file_to_directory

    @move_file_to_directory.setter
    def move_file_to_directory(self, value):
        self._move_file_to_directory = value

    @property
    def quote_strings(self):
        """
            bool: Controls the quoting of top-level rstrings. This parameter can be used only with the csv format. This parameter value is true by default.
        """
        return self._quote_strings

    @quote_strings.setter
    def quote_strings(self, value):
        self._quote_strings = value

    @property
    def separator(self):
        """
            str: Separator between records (defaults to comma ``,``).
        """
        return self._separator

    @separator.setter
    def separator(self, value):
        self._separator = value

    @property
    def suppress(self):
        """
            str: Specifies input attributes to be omitted from the output file. This parameter accepts one or more input attributes as values. Those named attributes are not output to the file. For example, you can use this parameter to omit file name inputs from the output file.
        """
        return self._suppress

    @suppress.setter
    def suppress(self, value):
        self._suppress = value

    @property
    def time_per_file(self):
        """
            float: Specifies the approximate time, in seconds, after which the current output file is closed and a new file is opened. If the :py:meth:`~streamsx.standard.files.FileSink.close_mode` parameter is set to time, this parameter must be specified.
        """
        return self._time_per_file

    @time_per_file.setter
    def time_per_file(self, value):
        self._time_per_file = value

    @property
    def truncate_on_reset(self):
        """
            bool: Specifies to truncate the file when a consistent region is reset.
        """
        return self._truncate_on_reset

    @truncate_on_reset.setter
    def truncate_on_reset(self, value):
        self._truncate_on_reset = value

    @property
    def tuples_per_file(self):
        """
            int: Specifies the maximum number of tuples that can be received for each output file. When the specified number of tuples are received, the current output file is closed and a new file is opened for writing. This parameter must be specified when the :py:meth:`~streamsx.standard.files.FileSink.close_mode` parameter is set to count.
        """
        return self._tuples_per_file

    @tuples_per_file.setter
    def tuples_per_file(self, value):
        self._tuples_per_file = value

    @property
    def write_failure_action(self):
        """
            enum: Specifies the action to take when file write fails. This parameter has values of ignore, log, and terminate of type :py:meth:`~streamsx.standard.WriteFailureAction`
        """
        return self._write_failure_action

    @write_failure_action.setter
    def write_failure_action(self, value):
        self._write_failure_action = value

    @property
    def write_punctuations(self):
        """
            bool: Specifies to write punctuations to the output file. It is false by default. This parameter can be used only with txt, csv, and bin formats.
        """
        return self._write_punctuations

    @write_punctuations.setter
    def write_punctuations(self, value):
        self._write_punctuations = value

    def populate(self, topology, stream, name, **options) -> streamsx.topology.topology.Sink:

        if self.append is not None:
            if self.append is True:
                self.append = streamsx.spl.op.Expression.expression('true')
        if self.bytes_per_file is not None:
            self.bytes_per_file = streamsx.spl.types.int32(self.bytes_per_file)
        if self.close_mode is not None:
            self.close_mode = streamsx.spl.op.Expression.expression(self.close_mode)
        if self.compression is not None:
            self.compression = streamsx.spl.op.Expression.expression(self.compression)
        if self.flush is not None:
            self.flush = streamsx.spl.types.uint32(self.flush)
        if self.flush_on_punctuation is not None:
            if self.flush_on_punctuation is True:
                self.flush_on_punctuation = streamsx.spl.op.Expression.expression('true')
            else:
                self.flush_on_punctuation = streamsx.spl.op.Expression.expression('false')
        if self.format is not None:
            self.format = streamsx.spl.op.Expression.expression(self.format)
        if self.has_delay_field is not None:
            if self.has_delay_field is True:
                self.has_delay_field = streamsx.spl.op.Expression.expression('true')
            else:
                self.has_delay_field = streamsx.spl.op.Expression.expression('false')
        if self.quote_strings is not None:
            if self.quote_strings is True:
                self.quote_strings = streamsx.spl.op.Expression.expression('true')
            else:
                self.quote_strings = streamsx.spl.op.Expression.expression('false')
        if self.suppress is not None:
            self.suppress = streamsx.spl.op.Expression.expression(self.suppress)
        if self.time_per_file is not None:
            self.time_per_file = streamsx.spl.types.float64(self.time_per_file)
        if self.truncate_on_reset is not None:
            if self.truncate_on_reset is True:
                self.truncate_on_reset = streamsx.spl.op.Expression.expression('true')
            else:
                self.truncate_on_reset = streamsx.spl.op.Expression.expression('false')
        if self.tuples_per_file is not None:
            self.tuples_per_file = streamsx.spl.types.int32(self.tuples_per_file)
        if self.write_failure_action is not None:
            self.write_failure_action = streamsx.spl.op.Expression.expression(self.write_failure_action)
        if self.write_punctuations is not None:
            if self.write_punctuations is True:
                self.write_punctuations = streamsx.spl.op.Expression.expression('true')
            else:
                self.write_punctuations = streamsx.spl.op.Expression.expression('false')

        _op = _FileSink(stream=stream, \
                        file=self.file, \
                        append=self.append, \
                        bytesPerFile=self.bytes_per_file, \
                        closeMode=self.close_mode, \
                        compression=self.compression, \
                        encoding=self.encoding, \
                        eolMarker=self.eol_marker, \
                        flush=self.flush, \
                        flushOnPunctuation=self.flush_on_punctuation, \
                        format=self.format, \
                        hasDelayField=self.has_delay_field, \
                        moveFileToDirectory=self.move_file_to_directory, \
                        quoteStrings=self.quote_strings, \
                        separator=self.separator, \
                        suppress=self.suppress, \
                        timePerFile=self.time_per_file, \
                        truncateOnReset=self.truncate_on_reset, \
                        tuplesPerFile=self.tuples_per_file, \
                        writeFailureAction=self.write_failure_action, \
                        writePunctuations=self.write_punctuations, \
                        name=name)

        return streamsx.topology.topology.Sink(_op)


def csv_reader(topology, schema, file, header=False, encoding=None, separator=None, ignoreExtraFields=False, hot=False, name=None):
    """Read a comma separated value file as a stream.

    The file defined by `file` is read and mapped to a stream
    with a structured schema of `schema`.

    Example for reading a file from application directory (file is part of application bundle)::

        import streamsx.standard.files as files
        from streamsx.topology.topology import Topology

        topo = Topology()
        sample_file = '/tmp/local/data.csv' # local file
        topo.add_file_dependency(sample_file, 'etc') # add sample file to etc dir in bundle
        fn = os.path.join('etc', 'data.csv') # file name relative to application dir
        sch = 'tuple<rstring a, int32 b>'
        r = files.csv_reader(topo, schema=sch, file=fn)

    Example for reading a file from file system accessible from the running job, for example persistent volume claim (Cloud Pak for Data)::

        import streamsx.standard.files as files
        from streamsx.topology.topology import Topology

        topo = Topology()
        sample_file = '/opt/ibm/streams-ext/data.csv' # file location accessible from running Streams application 
        r = files.csv_reader(topo, schema='tuple<rstring a, int32 b>', file=sample_file)

    Args:
        topology(Topology): Topology to contain the returned stream.
        schema(StreamSchema): Schema of the returned stream.
        file(str|Expression): Name of the source file. File name in relative path is expected in application directory, for example the file is added to the application bundle.
        header: Does the file contain a header.
        encoding: Specifies the character set encoding that is used in the output file.
        separator(str): Separator between records (defaults to comma ``,``).
        ignoreExtraFields:  When `True` then if the file contains more
            fields than `schema` has attributes they will be ignored.
            Otherwise if there are extra fields an error is raised.
        hot(bool): Specifies whether the input file is hot, which means it is appended continuously.
        name(str): Name of the stream, defaults to a generated name.

    Return:
        (Stream): Stream containing records from the file.
    """
    fe = streamsx.spl.op.Expression.expression(Format.csv.name)
    if isinstance(file, streamsx.spl.op.Expression) is False:
        if isinstance(file, str):
            if os.path.isabs(file) is False:       
                file = streamsx.spl.op.Expression.expression('getApplicationDir()+"'+'/'+file+'"')
                print("file="+str(file))
        else:
            raise TypeError(file)
    _op = _FileSource(topology, schema, file=file, format=fe, hotFile=hot,encoding=encoding,separator=separator,ignoreExtraCSVValues=ignoreExtraFields)
    return _op.outputs[0]


def csv_writer(stream, file, append=None, encoding=None, separator=None, flush=None, name=None):
    """Write a stream as a comma separated value file.

    The file defined by `file` is used as output file.

    Example for writing lines to a file::

        import streamsx.standard.files as files
        from streamsx.topology.topology import Topology

        topo = Topology()
        s = topo.source(range(13))
        sch = 'tuple<rstring a, int32 b>'
        s = s.map(lambda v: ('A'+str(v), v+7), schema=sch)
        files.csv_writer(s, file=os.path.join(tempfile.mkdtemp(), 'data.csv'))

    Args:
        topology(Topology): Topology to contain the returned stream.
        file(str|Expression): Name of the output file. File name in relative path is relative to data directory.
        append(bool): Specifies that the generated tuples are appended to the output file. If this parameter is false or not specified, the output file is truncated before the tuples are generated.
        encoding: Specifies the character set encoding that is used in the output file.
        separator(str): Separator between records (defaults to comma ``,``).
        flush(int): Specifies the number of tuples after which to flush the output file. By default no flushing on tuple numbers is performed. 
        name(str): Name of the stream, defaults to a generated name.

    Return:
        (streamsx.spl.op.Invoke): Sink operator

    """
    fe = streamsx.spl.op.Expression.expression(Format.csv.name)
    _op = _FileSink(stream, file, format=fe, append=append, encoding=encoding, separator=separator, flush=flush, name=name)


class _DirectoryScan(streamsx.spl.op.Source):
    def __init__(self, topology, schema,directory, pattern=None, sleepTime=None, initDelay=None, sortBy=None, order=None, moveToDirectory=None, ignoreDotFiles=None, ignoreExistingFilesAtStartup=None, name=None):
        kind="spl.adapter::DirectoryScan"
        inputs=None
        schemas=schema
        params = dict()
        params['directory'] = directory
        if pattern is not None:
            params['pattern'] = pattern
        if sleepTime is not None:
            params['sleepTime'] = sleepTime
        if initDelay is not None:
            params['initDelay'] = initDelay
        if sortBy is not None:
            params['sortBy'] = sortBy
        if order is not None:
            params['order'] = order
        if moveToDirectory is not None:
            params['moveToDirectory'] = moveToDirectory
        if ignoreDotFiles is not None:
            params['ignoreDotFiles'] = ignoreDotFiles
        if ignoreExistingFilesAtStartup is not None:
            params['ignoreExistingFilesAtStartup'] = ignoreExistingFilesAtStartup
        super(_DirectoryScan, self).__init__(topology,kind,schemas,params,name)


class _FileSource(streamsx.spl.op.Invoke):
    
    def __init__(self, topology, schemas, file=None, format=None, defaultTuple=None, parsing=None, hasDelayField=None, compression=None, eolMarker=None, blockSize=None, initDelay=None, hotFile=None, deleteFile=None, moveFileToDirectory=None, separator=None, encoding=None, hasHeaderLine=None, ignoreOpenErrors=None, readPunctuations=None, ignoreExtraCSVValues=None, name=None):
        kind="spl.adapter::FileSource"
        inputs=None
        params = dict()
        if file is not None:
            params['file'] = file
        if format is not None:
            params['format'] = format
        if defaultTuple is not None:
            params['defaultTuple'] = defaultTuple
        if parsing is not None:
            params['parsing'] = parsing
        if hasDelayField is not None:
            params['hasDelayField'] = hasDelayField
        if compression is not None:
            params['compression'] = compression
        if eolMarker is not None:
            params['eolMarker'] = eolMarker
        if blockSize is not None:
            params['blockSize'] = blockSize
        if initDelay is not None:
            params['initDelay'] = initDelay
        if hotFile is not None:
            params['hotFile'] = hotFile
        if deleteFile is not None:
            params['deleteFile'] = deleteFile
        if moveFileToDirectory is not None:
            params['moveFileToDirectory'] = moveFileToDirectory
        if separator is not None:
            params['separator'] = separator
        if encoding is not None:
            params['encoding'] = encoding
        if hasHeaderLine is not None:
            params['hasHeaderLine'] = hasHeaderLine
        if ignoreOpenErrors is not None:
            params['ignoreOpenErrors'] = ignoreOpenErrors
        if readPunctuations is not None:
            params['readPunctuations'] = readPunctuations
        if ignoreExtraCSVValues is not None:
            params['ignoreExtraCSVValues'] = ignoreExtraCSVValues
        super(_FileSource, self).__init__(topology,kind,inputs,schemas,params,name)
   

class _FileSink(streamsx.spl.op.Invoke):
    def __init__(self, stream, file, schema=None, format=None, flush=None, flushOnPunctuation=None, eolMarker=None, writePunctuations=None, hasDelayField=None, compression=None, separator=None, encoding=None, quoteStrings=None, closeMode=None, tuplesPerFile=None, timePerFile=None, bytesPerFile=None, moveFileToDirectory=None, append=None, writeFailureAction=None, suppress=None, truncateOnReset=None, writeStateHandlerCallbacks=None, name=None):
        topology = stream.topology
        kind="spl.adapter::FileSink"
        inputs=stream
        schemas=schema
        params = dict()
        params['file'] = file
        if format is not None:
            params['format'] = format
        if flush is not None:
            params['flush'] = flush
        if flushOnPunctuation is not None:
            params['flushOnPunctuation'] = flushOnPunctuation
        if eolMarker is not None:
            params['eolMarker'] = eolMarker
        if writePunctuations is not None:
            params['writePunctuations'] = writePunctuations
        if hasDelayField is not None:
            params['hasDelayField'] = hasDelayField
        if compression is not None:
            params['compression'] = compression
        if separator is not None:
            params['separator'] = separator
        if encoding is not None:
            params['encoding'] = encoding
        if quoteStrings is not None:
            params['quoteStrings'] = quoteStrings
        if closeMode is not None:
            params['closeMode'] = closeMode
        if tuplesPerFile is not None:
            params['tuplesPerFile'] = tuplesPerFile
        if timePerFile is not None:
            params['timePerFile'] = timePerFile
        if bytesPerFile is not None:
            params['bytesPerFile'] = bytesPerFile
        if moveFileToDirectory is not None:
            params['moveFileToDirectory'] = moveFileToDirectory
        if append is not None:
            params['append'] = append
        if writeFailureAction is not None:
            params['writeFailureAction'] = writeFailureAction
        if suppress is not None:
            params['suppress'] = suppress
        if truncateOnReset is not None:
            params['truncateOnReset'] = truncateOnReset
        if writeStateHandlerCallbacks is not None:
            params['writeStateHandlerCallbacks'] = writeStateHandlerCallbacks
        super(_FileSink, self).__init__(topology,kind,inputs,schema,params,name)



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017,2018
"""
Stream transformations using relational predicates.
"""

from streamsx.spl.op import Invoke, Map

import streamsx.standard._version
__version__ = streamsx.standard._version.__version__

class Aggregate(Map):
    """Aggregation against a window of a structured schema stream.

    The resuting stream (attribute ``stream``) will contain aggregations
    of the window defined by the methods invoked against the instance.

    The aggregation is invoked when the window is triggered.
    A grouped aggregation will result in `N` output tuples per aggregation
    where `N` is the number of groups in the window at the time of the
    trigger. A non-grouped window results in a single tuple per aggregation.
    A window punctuation mark is submitted after the tuples resulting
    from the aggregation.

    In all examples, an input schema of ``tuple<int32 id, timestamp ts, float64 reading>`` is assumed representing an input stream of sensor readings.

    Example of aggregating sensor readings to produce a stream containing
    the maximum, minimum and average reading over the last ten minutes
    updating every minute and grouped by sensor::

        from streamsx.standard.relational import Aggregate

        # Declare the window
        win = readings.last(datetime.timedelta(minutes=10)).trigger(datetime.timedelta(minutes=1))

        # Declare the output schema
        schema = 'tuple<int32 id, timestamp ts, float64 max_reading, float64 min_reading, float64 avg_reading>'


        # Invoke the aggregation
        agg = Aggregate.invoke(win, schema, group='id')

        # Declare the output attribute assignments.
        agg.min_reading = agg.min('reading')
        agg.max_reading = agg.max('reading')
        agg.avg_reading = agg.average('reading')

        # resulting stream is agg.stream
        agg.stream.print()

    When an output attribute is not assigned and it has a matching
    input attribute the value in the last tuple for the group is used.
    In the example above ``id`` will be set to the sensor identifier
    of the group and ``ts`` will be set to the timestamp of the most
    recent tuple for the group.

    For the output attribute assignment methods that return a grouped
    value, such as :py:meth:`count`, :py:meth:`max`, when the window is not
    grouped the value is against all tuples in the window.

    The aggregation is implemented using the ``spl.relational::Aggregate``
    SPL primitive operator from the SPL Standard toolkit.
    """
    @staticmethod
    def invoke(window, schema, group=None, name=None):
        """Invoke an aggregation against a window.

        Args:
            window(streamsx.topology.topology.Window): Window to aggregate against.
            schema(str,StreamSchema): Schema of output stream containing aggregations.
            group(str): Attribute name to group aggregations.
            name(str): Invocation name, defaults to a generated name.

        Returns:
             Aggregate: Aggregate invocation.
        """
        _op = Aggregate(window, schema, group=group, name=name)
        return _op
  
    def _output_func(self, name, attribute=None):
        _eofn = name + '('
        if attribute is not None:
            _eofn = _eofn + attribute
        _eofn = _eofn + ')'
        return self.output(self.expression(_eofn))
        
    def __init__(self, window, schema, group=None, partition=None, aggregateIncompleteWindows=None, aggregateEvictedPartitions=None, name=None):
        topology = window.topology
        kind="spl.relational::Aggregate"
        params = dict()
        if group is not None:
            params['groupBy'] = group
        if partition is not None:
            params['partitionBy'] = partition
        if aggregateIncompleteWindows is not None:
            params['aggregateIncompleteWindows'] = aggregateIncompleteWindows
        if aggregateEvictedPartitions is not None:
            params['aggregateEvictedPartitions'] = aggregateEvictedPartitions
        super(Aggregate, self).__init__(kind,window,schema,params,name)

    def count(self):
        """Count of tuples in the group.

        Returns an output expression of type ``int32`` representing
        the number of tuples in the group.

        Example::

            # Count the number of tuples grouped by sensor id in the last minute
            schema = 'tuple<int32 id, timestamp ts, int32 n>'
            agg = Aggregate.invoke(s.last(datetime.timedelta(minutes=1)), schema, group='id')
            agg.n = agg.count()

        Returns:
            Expression: Output expression with the type ``int32``.
        """
        return self._output_func('Count')

    def count_all(self):
        """Count of all tuples in the window.

        Returns:
            Expression: Output expression with the type ``int32``.
        """
        return self._output_func('CountAll')
    def count_groups(self):
        """Count of all groups in the window.

        Returns:
            Expression: Output expression with the type ``int32``.
        """
        return self._output_func('CountGroups')

    def max(self, attribute):
        """Maximum value for an input attribute.

        Returns an output expression representing the maximum value
        of the input attribute in the group.

        Example::

            # Get the maximum reading of the last ten tuples grouped by sensor id in the last minute
            schema = 'tuple<int32 id, timestamp ts, float64 max_reading>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.max_reading = agg.max('reading')

        Args:
            attribute(str): Attribute name to find the maximum value of.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('Max', attribute)

    def min(self, attribute):
        """Minimum value for an input attribute.

        Returns an output expression representing the minimum value
        of the input attribute in the group.

        Example::

            # Get the minimum reading of the last ten tuples grouped by sensor id
            # updating every input tuple.
            schema = 'tuple<int32 id, timestamp ts, float64 min_reading>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.min_reading = agg.min('reading')

        Args:
            attribute(str): Attribute name to find the minimum value of.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('Min', attribute)

    def sum(self, attribute):
        """Sum of values for an input attribute.

        Returns an output expression representing the sum of values
        of the input attribute in the group.

        Example::

            # Get the sum reading of the last ten tuples grouped by sensor id
            # updating every input tuple.
            schema = 'tuple<int32 id, timestamp ts, float64 sum_readings>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.sum_readings = agg.sum('reading')

        Args:
            attribute(str): Attribute name to find the sum of.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('Sum', attribute)

    def average(self, attribute):
        """Average of values for an input attribute.

        Returns an output expression representing the average of values
        of the input attribute in the group.

        Example::

            # Get the average reading of the last ten tuples grouped by sensor id
            # updating every input tuple.
            schema = 'tuple<int32 id, timestamp ts, float64 avg_reading>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.avg_reading = agg.average('reading')

        Args:
            attribute(str): Attribute name to find the average value of.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('Average', attribute)

    def first(self, attribute):
        """First value for an input attribute.

        Returns an output expression representing the first (earliest) value
        of the input attribute in the group.

        Example::

            # Get the first reading of the last ten tuples grouped by sensor id
            # updating every input tuple.
            schema = 'tuple<int32 id, timestamp ts, float64 first_reading>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.first_reading = agg.first('reading')

        Args:
            attribute(str): Attribute name to find the first value of.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('First', attribute)

    def last(self, attribute):
        """Last value for an input attribute.

        Returns an output expression representing the last
        (latest, most recent) value of the input attribute in the group.

        Example::

            # Get the last reading of the last ten tuples grouped by sensor id
            # updating every input tuple.
            schema = 'tuple<int32 id, timestamp ts, float64 last_reading>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.last_reading = agg.last('reading')

        Args:
            attribute(str): Attribute name to find the first value of.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('Last', attribute)

    def std(self, attribute, sample=False):
        """Standard deviation of values for an input attribute.

        Returns an output expression representing the standard deviation of values
        of the input attribute in the group.

        Example::

            # Get the standard deviation reading of the last ten tuples grouped by sensor id
            # updating every input tuple.
            schema = 'tuple<int32 id, timestamp ts, float64 sd_reading>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.sd_reading = agg.standard_('reading')

        Args:
            attribute(str): Attribute name to find the average value of.
            sample(bool): True to use sample standard deviation otherwise population standard deviation is used.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('SampleStdDev' if sample else 'PopulationStdDev', attribute)


class _Filter(Invoke):
    @staticmethod
    def matching(stream, filter, name=None):
        _op = Filter(stream, name=name)
        _op.params['filter'] = _op.expression(filter);
        return _op.outputs[0]

    def __init__(self, stream, filter=None, non_matching=False, name=None):
        topology = stream.topology
        kind="spl.relational::Filter"
        inputs=stream
        schema = stream.oport.schema
        schemas = [schema,schema] if non_matching else schema
        params = dict()
        if filter is not None:
            params['filter'] = filter
        super(Filter, self).__init__(topology,kind,inputs,schemas,params,name)


class Functor(Invoke):
    """Transform input tuples into output ones, and optionally filter them.

    If you do not filter an input tuple, any incoming tuple results in a tuple on each output stream

    The schema transformation is implemented using the ``spl.relational::Functor``
    SPL primitive operator from the SPL Standard toolkit.

    Example with schema transformation and two output streams::

        topo = Topology()
        s = U.sequence(topo, iterations=10) # schema is 'tuple<uint64 seq, timestamp ts>'
        fo = R.Functor.map(s, [StreamSchema('tuple<uint64 seq>'),StreamSchema('tuple<timestamp ts>')])
        seq_stream = fo.outputs[0] # schema is 'tuple<uint64 seq>' only
        ts_stream = fo.outputs[1] # schema is 'tuple<timestamp ts>' only

    Example with filter some tuples::

        topo = Topology()
        s = U.sequence(topo, iterations=5)
        fo = R.Functor.map(s, StreamSchema('tuple<uint64 seq>'), filter='seq>=2ul')
        fstream = fo.outputs[0]
        fstream.print()

    """
    @staticmethod
    def map(stream, schema, filter=None, name=None):
        """Map input stream schema to one or more output schemas

        Args:
            stream: Input stream
            schema(str,StreamSchema): Schema of output stream.
            filter(str): Attribute name to group aggregations.
            name(str): Invocation name, defaults to a generated name.

        Returns:
             Functor: Functor invocation.
        """
        _op = Functor(stream, schema, name=name)
        if filter is not None:
            _op.params['filter'] = _op.expression(filter);
        return _op
   
    def __init__(self, stream, schemas, filter=None, name=None):
        topology = stream.topology
        kind="spl.relational::Functor"
        inputs=stream
        params = dict()
        if filter is not None:
            params['filter'] = filter
        super(Functor, self).__init__(topology,kind,inputs,schemas,params,name)


class _Join(Invoke):
    @staticmethod
    def lookup(reference, reference_key, lookup, lookup_key, schema, name=None):
        _op = Join(reference, lookup.last(0), schemas=schema, name=name)
        _op.params['equalityLHS'] = _op.attribute(reference.stream, reference_key)
        _op.params['equalityRHS'] = _op.attribute(lookup, lookup_key)
        return _op.outputs[0]

    def __init__(self, left, right, schemas, match=None, algorithm=None, defaultTupleLHS=None, defaultTupleRHS=None, equalityLHS=None, equalityRHS=None, partitionByLHS=None, partitionByRHS=None, name=None):
        topology = left.topology
        kind="spl.relational::Join"
        inputs = [left, right]
        params = dict()
        if match is not None:
            params['match'] = match
        if algorithm is not None:
            params['algorithm'] = algorithm
        if defaultTupleLHS is not None:
            params['defaultTupleLHS'] = defaultTupleLHS
        if defaultTupleRHS is not None:
            params['defaultTupleRHS'] = defaultTupleRHS
        if equalityLHS is not None:
            params['equalityLHS'] = equalityLHS
        if equalityRHS is not None:
            params['equalityRHS'] = equalityRHS
        if partitionByLHS is not None:
            params['partitionByLHS'] = partitionByLHS
        if partitionByRHS is not None:
            params['partitionByRHS'] = partitionByRHS
        super(Join, self).__init__(topology,kind,inputs,schemas,params,name)



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017

"""
Functionality common to multiple modules.
"""

import enum

__all__ = ['CloseMode', 'WriteFailureAction', 'Format', 'Compression']

@enum.unique
class CloseMode(enum.Enum):
    """Write close modes."""
    punct = 0
    """Specifies to close the file when a window or final punctuation is received.."""
    count = 1
    """count is used with the tuples_per_file parameter to close the file when the specified number of tuples have been received."""
    size = 2
    """size is used with the bytes_per_file parameter to close the file when the specified number of bytes have been received.."""
    time = 3
    """time is used with the time_per_file parameter to close the file when the specified time has elapsed."""
    dynamic = 4
    """The file parameter can reference input attributes and is evaluated at each input tuple to compute the file name. If the file name is different from the previous value, the output file closes, and a new file opens."""
    never = 5
    """Default close mode."""

@enum.unique
class WriteFailureAction(enum.Enum):
    """Write failure actions."""
    ignore = 0
    """No action is taken on a write failure, and all future writes fail as well."""
    log = 1
    """The error is logged, and the error condition is cleared."""
    terminate = 2
    """The error is logged, and the operator terminates."""

@enum.unique
class Format(enum.Enum):
    """Formats for adapters."""
    csv = 0
    """Comma separated values."""
    txt = 1
    """Streams character representation of a tuple."""
    bin = 2
    """Streams binary representation of a tuple."""
    block = 3
    """Block of binary data."""
    line = 4
    """Line of text."""

@enum.unique
class Compression(enum.Enum):
    """Supported data compression algorithms."""
    zlib = 0
    """`zlib <https:://en.wikipedia.org/wiki/Zlib>`_ data compression."""
    gzip = 1
    """`gzip <https:://en.wikipedia.org/wiki/Gzip>`_ data compression."""
    bzip2 = 2
    """`bzip2 <https:://en.wikipedia.org/wiki/Bzip2>`_ data compression."""



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017,2018
"""
Standard utilities for processing streams.
"""

import streamsx.spl.op
from streamsx.topology.schema import StreamSchema
from streamsx.spl.types import float64, uint32, uint64

import streamsx.standard._version
__version__ = streamsx.standard._version.__version__

SEQUENCE_SCHEMA = StreamSchema('tuple<uint64 seq, timestamp ts>')
"""Structured schema containing a sequence identifier and a timestamp.

``'tuple<uint64 seq, timestamp ts>'``
"""

def sequence(topology, period=None, iterations=None, delay=None, name=None):
    """Create a sequence stream.

    Creates a structured stream with schema :py:const:`SEQUENCE_SCHEMA` with
    the ``seq`` attribute starting at zero and monotonically increasing and
    ``ts`` attribute set to the time the tuple was generated.

    Args:
        period(float): Period of tuple generation in seconds, if `None` then tuples are generated as fast as possible.
        iterations(int): Number of tuples on the stream, if `None` then the stream is infinite.
        delay(float): Delay in seconds before the first tuple is submitted, if `None` then the tuples are submitted as soon as possible.
        name(str): Name of the stream, if `None` a generated name is used.

    Returns:
        Stream: Structured stream containing an ever increasing ``seq`` attribute.
    """
    if iterations is not None:
        iterations = int(iterations)
    if period is not None:
        period = float(period)
    if name is None:
        name = 'Sequence'
        if iterations is not None:
            name = name + '({:d})'.format(iterations)
        if period is not None:
            name = name + ':period={:.3f}s'.format(period)

    _op = _Beacon(topology, SEQUENCE_SCHEMA, period=period, iterations=iterations, delay=delay, name=name)
    _op.seq = _op.output('IterationCount()')
    _op.ts = _op.output('getTimestamp()')
    return _op.stream

class _Beacon(streamsx.spl.op.Source):
    def __init__(self, topology, schema, period=None, iterations=None, delay=None, triggerCount=None, name=None):
        kind="spl.utility::Beacon"
        inputs=None
        schemas=schema
        params = dict()
        if period is not None:
            params['period'] = float64(period)
        if iterations is not None:
            params['iterations'] = uint32(iterations)
        if delay is not None:
            params['initDelay'] = float64(delay)
        if triggerCount is not None:
            params['triggerCount'] = triggerCount
        super(_Beacon, self).__init__(topology,kind,schemas,params,name)


def spray(stream, count, queue=1000, name=None):
    """Spray tuples to a number of streams.
    Each tuple on `stream` is sent to one (and only one)
    of the returned streams.
    The stream for a specific tuple is not defined,
    instead each stream has a dedicated thread and the
    first available thread will take the tuple and
    submit it.

    Each tuple on `stream` is placed on internal queue before it
    is submitted to an output stream. If the queue fills up
    then processing of the input stream is blocked until there
    is space in the queue.

    Args:
        count(int): Number of output streams the input stream will be sprayed across.
        queue(int): Maximum queue size.
        name(str): Name of the stream, if `None` a generated name is used.

    Returns:
        list(Stream) : List of output streams.
    """
    _op = _ThreadedSplit(stream, count, queue,name=name)
    return _op.outputs


class _ThreadedSplit (streamsx.spl.op.Invoke):
    def __init__(self, stream, count, queue=1000, name=None):
        topology = stream.topology
        kind="spl.utility::ThreadedSplit"
        inputs=stream
        schemas=[stream.oport.schema] * count
        params = dict()
        params['bufferSize'] = uint32(queue)
        super(_ThreadedSplit, self).__init__(topology,kind,inputs,schemas,params,name)


def throttle(stream, rate, precise=False, name=None):
    """Throttle the rate of a stream.

    Args:
         rate(float): Throttled rate of the returned stream in tuples/second.
         precise(bool): Try to make the rate precise at the cost of increased overhead.
         name(str): Name of the stream, if `None` a generated name is used.
    """
    _op = _Throttle(stream, rate, precise=precise, name=name)
    return _op.stream

class _Throttle (streamsx.spl.op.Map):
    """Stream throttle capability
    """
    def __init__(self, stream, rate, period=None, includePunctuations=None, precise=None, name=None):
        kind="spl.utility::Throttle"
        params = dict()
        params['rate'] = float64(rate)
        if period is not None:
            params['period'] = float64(period)
        if includePunctuations is not None:
            params['includePunctuations'] = includePunctuations
        if precise is not None:
            params['precise'] = precise
        super(_Throttle, self).__init__(kind,stream,params=params,name=name)


def union(inputs, schema, name=None):
    """Union structured streams with disparate schemas.

    Each tuple on any of the streams in `inputs` results in
    a tuple on the returned stream.

    All attributes of the output tuple are set from the input tuple,
    thus the schema of each input must include attributes matching
    (name and type) the output schema.

    The order of attributes in the input schemas need not match
    the output schemas and the input schemas may contain additional
    attributes which will be discarded.

    .. note:: 
        This method differs from ``Stream.union`` in that 
        the schemas of input and output streams can differ, while
        ``Stream.union`` requires matching input and output schemas.

    Args:
        inputs(list[Stream]): Streams to be unioned.
        schema(StreamSchema): Schema of output stream
        name(str): Name of the stream, if `None` a generated name is used.

    Returns:
        Stream: Stream that is a union of `inputs`.

    """
    _op = _Union(inputs, schema, name=name)
    return _op.outputs[0]

class _Union (streamsx.spl.op.Invoke):
    """Union structured streams with disparate schemas.
    """

    def __init__(self, inputs, schema, name=None):
        topology = inputs[0].topology
        kind="spl.utility::Union"
        schemas=schema
        params = None
        super(_Union, self).__init__(topology,kind,inputs,schemas,params,name)


def deduplicate(stream, count=None, period=None, name=None):
    """Deduplicate tuples on a stream.

    If a tuple on `stream` is followed by a duplicate tuple
    within `count` tuples or `period` number of seconds
    then the duplicate is discarded from the returned stream.

    Only one of `count` or `period` can be set.

    Args:
        stream(Stream): Stream to be deduplicated.
        count(int): Number of tuples.
        period(float): Time period to check for duplicates.
        name(str): Name of resultant stream, defaults to a generated name.

    Returns:
        Stream: Deduplicated stream.
    """
    if count and period:
        raise ValueError("Cannot set count and period")

    _op = _DeDuplicate(stream, count=count, period=period, name=name)
    return _op.stream

class _DeDuplicate (streamsx.spl.op.Map):
    def __init__(self, stream, timeOut=None, count=None, deltaAttribute=None, delta=None, key=None, resetOnDuplicate=None, flushOnPunctuation=None, name=None):
        kind="spl.utility::DeDuplicate"
        params = dict()
        if timeOut is not None:
            params['timeOut'] = float64(timeOut)
        if count is not None:
            params['count'] = int(count)
        if deltaAttribute is not None:
            params['deltaAttribute'] = deltaAttribute
        if delta is not None:
            params['delta'] = delta
        if key is not None:
            params['key'] = key
        if resetOnDuplicate is not None:
            params['resetOnDuplicate'] = resetOnDuplicate
        if flushOnPunctuation is not None:
            params['flushOnPunctuation'] = flushOnPunctuation
        super(_DeDuplicate, self).__init__(kind,stream,params=params,name=name)

def delay(stream, delay, max_delayed=1000, name=None):
    """Delay tuples on a stream.

    Delays tuples on `stream` maintaining inter-arrival times
    of tuples and punctuation.

    Example delaying a stream ``ss`` by 0.5 seconds::

        import streamsx.standard.utility as U

        readings = U.delay(readings, delay=1.5)

    Args:
        stream(Stream): Stream to be delayed.
        delay(float): Seconds to delay each tuple.
        max_delayed(int): Number of items that can be delayed before upstream processing is blocked.
        name(str): Name of resultant stream, defaults to a generated name.

    Returns:
        Stream: Delayed stream.
    """
    _op = _Delay(stream, delay, max_delayed, name)
    return _op.stream

class _Delay(streamsx.spl.op.Map):
    def __init__(self, stream, delay, max_delayed=1000, name=None):
        topology = stream.topology
        kind="spl.utility::Delay"
        params = dict()
        params['delay'] = float64(delay)
        if max_delayed is not None:
            params['bufferSize'] = uint32(max_delayed)
        super(_Delay, self).__init__(kind,stream,params=params,name=name)

def pair(stream0, stream1, matching=None, name=None):
    """Pair tuples across two streams.

    This method is used to merge results from performing
    parallel tasks on the same stream, for example peform multiple
    model scoring on the same stream.

    Holds tuples on the two input streams until a matched tuple has been
    received by both input streams. Once matching tuples have received
    the two tuples are submitted to the returned stream with the
    tuple from ``stream0`` followed by the one from ``stream1``.

    Tuples are matched according to the ``matching`` parameter which
    is an attribute name from the input tuple schema,
    typically representing the application key of the tuple.

    If ``matching`` is ``None`` then a match occurs when
    a tuple is received, so that tuples are emitted when a tuple
    has been received by both input streams.
    
    ``stream0`` and ``stream1`` must have the same schema and the resultant
    stream has the same schema.

    These schemas are not supported when ``matching`` is specified.

       * ``CommonSchema.Python``
       * ``CommonSchema.Json``

    This is equivalent to ``merge([stream0, stream1], matching, name)``.

    Example of scoring in parallel::

        import streamsx.standard.utility as U
        
        # Stream of customer information with customer identifier
        # as the id attribute.
        customers = ...
        score_schema = schema.extend(StreamSchema('tuple<float64 score>'))
    
        # Score each tuple on customers in parallel
        cust_churn = s.map(customer_churn_score, schema=score_schema)
        cust_renew = s.map(customer_renew_score, schema=score_schema)
        
        # Pair back as single stream
        # cust_churn_renew stream will contain two tuples for
        # each customer, the churn score followed by the renew score.
        cust_churn_renew = U.pair(cust_churn, cust_renew, matching='id');

    Args:
        stream0(Stream): First input stream.
        stream1(Stream): Second input stream.
        matching(str): Attribute name for matching tuples.
        name(str): Name of resultant stream, defaults to a generated name.

    Returns:
        Stream: Paired stream.
    """
    return merge([stream0, stream1], matching, name)

def merge(inputs, matching=None, name=None):
    """Merge tuples across two (or more) streams.

    This method is used to merge results from performing
    parallel tasks on the same stream, for example peform multiple
    model scoring on the same stream.

    Holds tuples on the input streams until a matched tuple has been
    received by each input stream. Once matching tuples have received
    for all input streams the tuples are submitted to the returned
    stream in order of the input ports.

    Tuples are matched according to the ``matching`` parameter which
    is an attribute name from the input tuple schema,
    typically representing the application key of the tuple.

    If ``matching`` is ``None`` then a match occurs when
    a tuple is received, so that tuples are emitted when a tuple
    has been received by each input port.
    
    All input streams must have the same schema and the resultant
    stream has the same schema.

    These schemas are not supported when ``matching`` is specified.

       * ``CommonSchema.Python``
       * ``CommonSchema.Json``

    Args:
        inputs(list[Stream]): Input streams to be matched.
        matching(str): Attribute name for matching.
        name(str): Name of resultant stream, defaults to a generated name.

    Returns:
        Stream: Merged stream.
    """
    _op = _Pair(inputs, matching, name=name)
    return _op.outputs[0]

class _Pair(streamsx.spl.op.Invoke):
    def __init__(self, inputs, matching=None, buffer_size=None, name=None):
        topology = inputs[0].topology
        kind="spl.utility::Pair"
        schema=inputs[0].oport.schema
        params = dict()
        if buffer_size is not None:
            params['bufferSize'] = uint32(buffer_size)
        super(_Pair, self).__init__(topology,kind,inputs,[schema],params,name)
        if matching is not None:
            for port_idx in range(len(inputs)):
                self.params['partitionBy' + str(port_idx)] = self.attribute(inputs[port_idx], matching)

def gate(stream, control, max_unacked=1, ack_count=1, name=None):
    """Gate tuple flow based upon downstream processing.

    Tuples on `stream` are passed through unmodified to the returned
    stream but the flow is gated by tuples on the `control` stream.

    Up to `max_unacked` tuples flow through the gate before `stream` is
    is blocked. Each tuple arriving on `control` acknowledges `ack_count`
    tuples and thus unblocks `stream` until the number of unacknowledged
    tuples reaches `max_unacked` again.

    The output of some downstream processing is typically used as `control`
    and thus `control` is usually a stream obtained from a `streamsx.topology.topology.PendingStream`.

    Args:
        stream(Stream): Stream to be gated.
        control(Stream): Controlling stream.
        max_unacked(int): Maximum of tuples allowed through the gate without acknowledgement.
        ack_count(int): Count of tuples to acknowledge with each tuple arriving on `control`.
        name(str): Name of resultant stream, defaults to a generated name.
    
    Returns:
        Stream: Gated stream.
    """
    ack_count = uint32(ack_count)
    _op = _Gate([stream,control], maxUnackedTupleCount=max_unacked, numTuplesToAck=ack_count,name=name)
    return _op.outputs[0]

class _Gate(streamsx.spl.op.Invoke):
    def __init__(self, inputs, maxUnackedTupleCount, numTuplesToAck=None, name=None):
        topology = inputs[0].topology
        kind="spl.utility::Gate"
        schema=inputs[0].oport.schema
        params = dict()
        if maxUnackedTupleCount is not None:
            params['maxUnackedTupleCount'] = uint32(maxUnackedTupleCount)
        if numTuplesToAck is not None:
            params['numTuplesToAck'] = numTuplesToAck
        super(_Gate, self).__init__(topology,kind,inputs,[schema],params,name)



# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017

import streamsx.spl.op
from streamsx.topology.schema import StreamSchema

import enum
Format = enum.Enum('Format', 'csv txt bin block line')
Compression = enum.Enum('Compression', 'zlib gzip bzip2')
CloseMode = enum.Enum('CloseMode', 'punct count size time dynamic never')
WriteFailureAction = enum.Enum('WriteFailureAction', 'ignore log terminate')

